# -*- coding: utf-8 -*-
"""
:Module:         salespyforce.utils.tests.test_instantiate_object
:Synopsis:       This module is used by pytest to test instantiating the core object
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  29 May 2023
"""

from . import resources


def test_instantiate_core_object():
    """This function tests the ability to instantiate the core object.

    .. versionadded:: 1.1.0
    """
    sfdc_object = resources.get_core_object()
    assert 'force.com' in sfdc_object.base_url


def test_get_api_versions():
    """This function tests the get_api_versions() method in the core object.

    .. versionadded:: 1.1.0
    """
    sfdc_object = resources.get_core_object()
    api_versions = sfdc_object.get_api_versions()
    assert isinstance(api_versions, list) and 'version' in api_versions[0]


def test_get_rest_resources():
    """This function tests the get_rest_resources() method in the core object.

    .. versionadded:: 1.1.0
    """
    sfdc_object = resources.get_core_object()
    rest_resources = sfdc_object.get_rest_resources()
    assert 'metadata' in rest_resources


def test_get_org_limits():
    """This function tests the get_org_limits() method in the core object.

    .. versionadded:: 1.1.0
    """
    sfdc_object = resources.get_core_object()
    org_limits = sfdc_object.get_org_limits()
    assert 'DailyApiRequests' in org_limits



# -*- coding: utf-8 -*-
"""
:Module:         salespyforce.utils.tests.test_sobjects
:Synopsis:       This module is used by pytest to test basic sObject-related methods
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  29 May 2023
"""

import requests

from . import resources


def test_get_all_sobjects():
    """This function tests the get_all_sobjects() method in the core object.

    .. versionadded:: 1.1.0
    """
    sfdc_object = resources.get_core_object()
    all_sobjects = sfdc_object.get_all_sobjects()
    assert 'sobjects' in all_sobjects


def test_get_and_describe_sobject():
    """This function tests the get_sobject() and describe_object() methods in the core object.

    .. versionadded:: 1.1.0
    """
    # Instantiate the core object
    sfdc_object = resources.get_core_object()

    # Test the default query (non-describe)
    account_sobject = sfdc_object.get_sobject('Account')
    assert 'objectDescribe' in account_sobject

    # Test with describe enabled
    account_sobject_describe = sfdc_object.get_sobject('Account', describe=True)
    assert 'activateable' in account_sobject_describe

    # Test the describe_object() method
    account_describe = sfdc_object.describe_object('Account')
    assert 'activateable' in account_describe


def test_create_record(monkeypatch):
    # Instantiate the core object
    sfdc_object = resources.get_core_object()

    # Overwrite the requests.post functionality with the mock_success_post() function
    monkeypatch.setattr(requests, 'post', resources.mock_success_post)

    # Perform the mock API call
    payload = {
        "Name": "Express Logistics and Transport"
    }
    response = sfdc_object.create_sobject_record('Account', payload)
    assert 'success' in response and response.get('success') is True and 'id' in response



# -*- coding: utf-8 -*-
"""
:Package:           salespyforce.errors
:Synopsis:          This module includes custom exceptions and handlers
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     13 Mar 2023
"""

__all__ = ['exceptions', 'handlers']

from . import exceptions, handlers



# -*- coding: utf-8 -*-
"""
:Module:            salespyforce.api
:Synopsis:          Defines the basic functions associated with the Salesforce API
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     17 Feb 2023
"""

import requests

from .utils import log_utils

# Initialize logging
logger = log_utils.initialize_logging(__name__)


def get(sfdc_object, endpoint, params=None, headers=None, timeout=30, show_full_error=True, return_json=True):
    """This method performs a GET request against the Salesforce instance.
    (`Reference <https://jereze.com/code/authentification-salesforce-rest-api-python/>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :param endpoint: The API endpoint to query
    :type endpoint: str
    :param params: The query parameters (where applicable)
    :type params: dict, None
    :param headers: Specific API headers to use when performing the API call
    :type headers: dict, None
    :param timeout: The timeout period in seconds (defaults to ``30``)
    :type timeout: int, str, None
    :param show_full_error: Determines if the full error message should be displayed (defaults to ``True``)
    :type show_full_error: bool
    :param return_json: Determines if the response should be returned in JSON format (defaults to ``True``)
    :returns: The API response in JSON format or as a ``requests`` object
    """
    # Define the parameters as an empty dictionary if none are provided
    params = {} if params is None else params

    # Define the headers
    default_headers = _get_headers(sfdc_object.access_token)
    headers = default_headers if not headers else headers

    # Make sure the endpoint begins with a slash
    endpoint = f'/{endpoint}' if not endpoint.startswith('/') else endpoint

    # Perform the API call
    response = requests.get(f'{sfdc_object.instance_url}{endpoint}', headers=headers, params=params, timeout=timeout)
    if response.status_code >= 300:
        if show_full_error:
            raise RuntimeError(f'The GET request failed with a {response.status_code} status code.\n'
                               f'{response.text}')
        else:
            raise RuntimeError(f'The GET request failed with a {response.status_code} status code.')
    if return_json:
        response = response.json()
    return response


def api_call_with_payload(sfdc_object, method, endpoint, payload, params=None, headers=None, timeout=30,
                          show_full_error=True, return_json=True):
    """This method performs a POST call against the Salesforce instance.
    (`Reference <https://jereze.com/code/authentification-salesforce-rest-api-python/>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :param method: The API method (``post``, ``put``, or ``patch``)
    :type method: str
    :param endpoint: The API endpoint to query
    :type endpoint: str
    :param payload: The payload to leverage in the API call
    :type payload: dict
    :param params: The query parameters (where applicable)
    :type params: dict, None
    :param headers: Specific API headers to use when performing the API call
    :type headers: dict, None
    :param timeout: The timeout period in seconds (defaults to ``30``)
    :type timeout: int, str, None
    :param show_full_error: Determines if the full error message should be displayed (defaults to ``True``)
    :type show_full_error: bool
    :param return_json: Determines if the response should be returned in JSON format (defaults to ``True``)
    :returns: The API response in JSON format or as a ``requests`` object
    """
    # Define the parameters as an empty dictionary if none are provided
    params = {} if params is None else params

    # Define the headers
    default_headers = _get_headers(sfdc_object.access_token)
    headers = default_headers if not headers else headers

    # Make sure the endpoint begins with a slash
    endpoint = f'/{endpoint}' if not endpoint.startswith('/') else endpoint

    # Perform the API call
    if method.lower() == 'post':
        response = requests.post(f'{sfdc_object.instance_url}{endpoint}', json=payload, headers=headers, params=params,
                                 timeout=timeout)
    elif method.lower() == 'patch':
        response = requests.patch(f'{sfdc_object.instance_url}{endpoint}', json=payload, headers=headers, params=params,
                                  timeout=timeout)
    elif method.lower() == 'put':
        response = requests.put(f'{sfdc_object.instance_url}{endpoint}', json=payload, headers=headers, params=params,
                                timeout=timeout)
    else:
        raise ValueError('The API call method (POST or PATCH OR PUT) must be defined.')

    # Examine the result
    if response.status_code >= 300:
        if show_full_error:
            raise RuntimeError(f'The POST request failed with a {response.status_code} status code.\n'
                               f'{response.text}')
        else:
            raise RuntimeError(f'The POST request failed with a {response.status_code} status code.')
    if return_json:
        try:
            response = response.json()
        except Exception as exc:
            print(f'Failed to convert the API response to JSON format due to the following exception: {exc}')
    return response


def _get_headers(_access_token, _header_type='default'):
    """This function returns the appropriate HTTP headers to use for different types of API calls."""
    headers = {
        'content-type': 'application/json',
        'accept-encoding': 'gzip',
        'authorization': f'Bearer {_access_token}'
    }
    if _header_type == 'articles':
        headers['accept-language'] = 'en-US'
    return headers



# -*- coding: utf-8 -*-
"""
:Module:         salespyforce.utils.tests.test_sobjects
:Synopsis:       This module is used by pytest to test performing SOQL queries
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  29 May 2023
"""

import requests

from . import resources


def test_search_string(monkeypatch):
    """This function tests the ability to search for a string using an SOSL query.

    .. versionadded:: 1.1.0
    """
    # Instantiate the core object
    sfdc_object = resources.get_core_object()

    # Overwrite the requests.post functionality with the mock_success_post() function
    monkeypatch.setattr(requests, 'get', resources.mock_sosl_get)

    # Perform the mock API call
    result = sfdc_object.search_string('Account')
    assert 'searchRecords' in result




# -*- coding: utf-8 -*-
"""
:Module:            salespyforce
:Synopsis:          This is the ``__init__`` module for the salespyforce package
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     08 May 2023
"""

from . import core
from .core import Salesforce
from .utils import version

__all__ = ['core', 'Salesforce']

# Define the package version by pulling from the highspot.utils.version module
__version__ = version.get_full_version()


# Allow the core.define_connection_info() function to be executed directly
def define_connection_info():
    """This function prompts the user for the connection information.

    :returns: The connection info in a dictionary
    """
    return core.define_connection_info()


# Allow the core.compile_connection_info() function to be executed directly
def compile_connection_info(base_url, org_id, username, password, endpoint_url,
                            client_id, client_secret, security_token):
    """This function compiles the connection info into a dictionary that can be consumed by the core object.

    :param base_url: The base URL of the Salesforce instance
    :type base_url: str
    :param org_id: The Org ID of the Salesforce instance
    :type org_id: str
    :param username: The username of the API user
    :type username: str
    :param password: The password of the API user
    :type password: str
    :param endpoint_url: The endpoint URL for the Salesforce instance
    :type endpoint_url: str
    :param client_id: The Client ID for the Salesforce instance
    :type client_id: str
    :param client_secret: The Client Secret for the Salesforce instance
    :type client_secret: str
    :param security_token: The Security Token for the Salesforce instance
    :type security_token: str
    :returns: The connection info in a dictionary
    """
    return core.compile_connection_info(base_url, org_id, username, password, endpoint_url,
                                        client_id, client_secret, security_token)



# -*- coding: utf-8 -*-
"""
:Module:            salespyforce.utils.helper
:Synopsis:          Module that allows the salespyforce library to leverage a helper configuration file
:Usage:             ``from salespyforce.utils import helper``
:Example:           ``helper_settings = helper.get_settings('/tmp/helper.yml', 'yaml')``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     13 Mar 2023
"""

import json

import yaml

from .. import errors
from . import log_utils
from .core_utils import get_file_type

# Initialize logging within the module
logger = log_utils.initialize_logging(__name__)


def import_helper_file(file_path, file_type):
    """This function imports a YAML (.yml) or JSON (.json) helper config file.

    :param file_path: The file path to the YAML file
    :type file_path: str
    :param file_type: Defines the file type as either ``yaml`` or ``json``
    :type file_type: str
    :returns: The parsed configuration data
    :raises: :py:exc:`FileNotFoundError`, :py:exc:`salespyforce.errors.exceptions.InvalidHelperFileTypeError`
    """
    with open(file_path, 'r') as cfg_file:
        if file_type == 'yaml':
            helper_cfg = yaml.safe_load(cfg_file)
        elif file_type == 'json':
            helper_cfg = json.load(cfg_file)
        else:
            raise errors.exceptions.InvalidHelperFileTypeError()
    logger.info(f'The helper file {file_path} was imported successfully.')
    return helper_cfg


def _convert_yaml_to_bool(_yaml_bool_value):
    """This function converts the 'yes' and 'no' YAML values to traditional Boolean values."""
    _true_values = ['yes', 'true']
    if _yaml_bool_value.lower() in _true_values:
        _bool_value = True
    else:
        _bool_value = False
    return _bool_value


def _get_connection_info(_helper_cfg):
    """This function parses any connection information found in the helper file."""
    _connection_info = {}
    _connection_keys = ['username', 'password', 'base_url', 'endpoint_url',
                        'client_key', 'client_secret', 'org_id', 'security_token']
    for _key in _connection_keys:
        if _key in _helper_cfg['connection']:
            _connection_info[_key] = _helper_cfg['connection'][_key]
    return _connection_info


def _collect_values(_top_level_keys, _helper_cfg, _helper_dict=None, _ignore_missing=False):
    """This function loops through a list of top-level keys to collect their corresponding values.

    :param _top_level_keys: One or more top-level keys that might be found in the helper config file
    :type _top_level_keys: list, tuple, set, str
    :param _helper_cfg: The configuration parsed from the helper configuration file
    :type _helper_cfg: dict
    :param _helper_dict: A predefined dictionary to which the key value pairs should be added
    :type _helper_dict: dict, None
    :param _ignore_missing: Indicates whether fields with null values should be ignored (``False`` by default)
    :type _ignore_missing: bool
    :returns: A dictionary with the identified key value pairs
    """
    _helper_dict = {} if not _helper_dict else _helper_dict
    _top_level_keys = (_top_level_keys, ) if isinstance(_top_level_keys, str) else _top_level_keys
    for _key in _top_level_keys:
        if _key in _helper_cfg:
            _key_val = _helper_cfg[_key]
            if _key_val in HelperParsing.yaml_boolean_values:
                _key_val = HelperParsing.yaml_boolean_values.get(_key_val)
            _helper_dict[_key] = _key_val
        elif _key == "ssl_verify":
            # Verify SSL certificates by default unless explicitly set to false
            _helper_dict[_key] = True
        else:
            if not _ignore_missing:
                _helper_dict[_key] = None
    return _helper_dict


def get_helper_settings(file_path, file_type='yaml', defined_settings=None):
    """This function returns a dictionary of the defined helper settings.

    :param file_path: The file path to the helper configuration file
    :type file_path: str
    :param file_type: Defines the helper configuration file as a ``yaml`` file (default) or a ``json`` file
    :type file_type: str
    :param defined_settings: Core object settings (if any) defined via the ``defined_settings`` parameter
    :type defined_settings: dict, None
    :returns: Dictionary of helper variables
    :raises: :py:exc:`salespyforce.errors.exceptions.InvalidHelperFileTypeError`
    """
    # Initialize the helper_settings dictionary
    helper_settings = {}

    # Convert the defined_settings parameter to an empty dictionary if null
    defined_settings = {} if not defined_settings else defined_settings

    if file_type != 'yaml' and file_type != 'json':
        file_type = get_file_type(file_path)

    # Import the helper configuration file
    helper_cfg = import_helper_file(file_path, file_type)

    # Populate the connection information in the helper dictionary
    if 'connection' in helper_cfg and 'connection' not in defined_settings:
        helper_settings['connection'] = _get_connection_info(helper_cfg)

    # Populate the SSL certificate verification setting in the helper dictionary
    if 'ssl_verify' not in defined_settings:
        helper_settings.update(_collect_values('ssl_verify', helper_cfg))

    # Return the helper_settings dictionary
    return helper_settings


class HelperParsing:
    """This class is used to help parse values imported from a YAML configuration file."""
    # Define dictionary to map YAML Boolean to Python Boolean
    yaml_boolean_values = {
        True: True,
        False: False,
        'yes': True,
        'no': False
    }



# -*- coding: utf-8 -*-
"""
:Module:            salespyforce.core
:Synopsis:          This module performs the core Salesforce-related operations
:Usage:             ``from salespyforce import Salesforce``
:Example:           ``sfdc = Salesforce()``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     22 Nov 2023
"""

import re

import requests

from . import api, errors
from . import chatter as chatter_module
from . import knowledge as knowledge_module
from .utils import core_utils, log_utils
from .utils.helper import get_helper_settings

# Define constants
CURRENT_SFDC_VERSION = '58.0'

# Initialize logging
logger = log_utils.initialize_logging(__name__)


class Salesforce(object):
    """This is the class for the core object leveraged in this module."""
    # Define the function that initializes the object instance (i.e. instantiates the object)
    def __init__(self, connection_info=None, version=CURRENT_SFDC_VERSION, base_url=None, org_id=None, username=None,
                 password=None, endpoint_url=None, client_id=None, client_secret=None, security_token=None, helper=None):
        """This method instantiates the core Salesforce object.

        :param connection_info: The information for connecting to the Salesforce instance
        :type connection_info: dict, None
        :param version: The Salesforce API version to utilize (Default: ``57.0``)
        :type version: str
        :param base_url: The base URL of the Salesforce instance
        :type base_url: str, None
        :param org_id: The Org ID of the Salesforce instance
        :type org_id: str, None
        :param username: The username of the API user
        :type username: str, None
        :param password: The password of the API user
        :type password: str, None
        :param endpoint_url: The endpoint URL for the Salesforce instance
        :type endpoint_url: str, None
        :param client_id: The Client ID for the Salesforce instance
        :type client_id: str, None
        :param client_secret: The Client Secret for the Salesforce instance
        :type client_secret: str, None
        :param security_token: The Security Token for the Salesforce instance
        :type security_token: str, None
        :param helper: The file path of a helper file
        :type helper: str, None
        :returns: The instantiated object
        :raises: :py:exc:`TypeError`
        """
        # Define the default settings
        self._helper_settings = {}

        # Check for provided connection info
        if connection_info is None:
            # Check for a supplied helper file
            if helper:
                # Parse the helper file contents
                self.helper_path = helper
                if any((isinstance(helper, tuple), isinstance(helper, list), isinstance(helper, set))):
                    helper_file_path, helper_file_type = helper
                elif isinstance(helper, str):
                    helper_file_path, helper_file_type = (helper, 'yaml')
                elif isinstance(helper, dict):
                    helper_file_path, helper_file_type = helper.values()
                else:
                    error_msg = "The 'helper' argument can only be supplied as tuple, string, list, set or dict."
                    logger.error(error_msg)
                    raise TypeError(error_msg)
                self._helper_settings = get_helper_settings(helper_file_path, helper_file_type)
                connection_info = self._parse_helper_connection_info()
            elif not any((base_url, org_id, username, password, endpoint_url, client_id, client_secret, security_token)):
                # Prompt for the connection info if not defined
                connection_info = define_connection_info()
            else:
                # Compile the connection info from the provided parameters
                connection_info = compile_connection_info(base_url, org_id, username, password, endpoint_url,
                                                          client_id, client_secret, security_token)

        # Get the connection information used to connect to the instance
        self.connection_info = connection_info if connection_info is not None else self._get_empty_connection_info()

        # Define the base URL value
        self.base_url = self.connection_info.get('base_url')

        # Define the version value
        self.version = f'v{version}'

        # Define the connection response data variables
        auth_response = self.connect()
        self.access_token = auth_response.get('access_token')
        self.instance_url = auth_response.get('instance_url')
        self.signature = auth_response.get('signature')

        # Import inner object classes so their methods can be called from the primary object
        self.chatter = self._import_chatter_class()
        self.knowledge = self._import_knowledge_class()

    def _import_chatter_class(self):
        """This method allows the :py:class:`salespyforce.core.Salesforce.Chatter` class to be utilized in the core object."""
        return Salesforce.Chatter(self)

    def _import_knowledge_class(self):
        """This method allows the :py:class:`salespyforce.core.Salesforce.Knowledge` class to be utilized in the core object."""
        return Salesforce.Knowledge(self)

    @staticmethod
    def _get_empty_connection_info():
        """This method returns an empty connection_info dictionary with all blank values."""
        _connection_info = {}
        _fields = ['username', 'password', 'base_url', 'endpoint_url',
                   'client_key', 'client_secret', 'org_id', 'security_token']
        for _field in _fields:
            _connection_info[_field] = ''
        return _connection_info

    def _parse_helper_connection_info(self):
        """This method parses the helper content to populate the connection info."""
        _connection_info = {}
        _fields = ['username', 'password', 'base_url', 'endpoint_url',
                   'client_key', 'client_secret', 'org_id', 'security_token']
        for _field in _fields:
            if _field in self._helper_settings['connection']:
                _connection_info[_field] = self._helper_settings['connection'][_field]
        return _connection_info

    def _get_headers(self, _header_type='default'):
        """This method returns the appropriate HTTP headers to use for different types of API calls."""
        return api._get_headers(_access_token=self.access_token, _header_type=_header_type)

    def connect(self):
        """This method connects to the Salesforce instance to obtain the access token.
        (`Reference <https://jereze.com/code/authentification-salesforce-rest-api-python/>`_)

        :returns: The API call response with the authorization information
        """
        params = {
            'grant_type': 'password',
            'client_id': self.connection_info.get('client_key'),
            'client_secret': self.connection_info.get('client_secret'),
            'username': self.connection_info.get('username'),
            'password': f'{self.connection_info.get("password")}{self.connection_info.get("security_token")}'
        }
        response = requests.post(self.connection_info.get('endpoint_url'), params=params)
        if response.status_code != 200:
            raise RuntimeError(f'Failed to connect to the Salesforce instance.\n{response.text}')
        return response.json()

    def get(self, endpoint, params=None, headers=None, timeout=30, show_full_error=True, return_json=True):
        """This method performs a GET request against the Salesforce instance.
        (`Reference <https://jereze.com/code/authentification-salesforce-rest-api-python/>`_)

        :param endpoint: The API endpoint to query
        :type endpoint: str
        :param params: The query parameters (where applicable)
        :type params: dict, None
        :param headers: Specific API headers to use when performing the API call
        :type headers: dict, None
        :param timeout: The timeout period in seconds (defaults to ``30``)
        :type timeout: int, str, None
        :param show_full_error: Determines if the full error message should be displayed (defaults to ``True``)
        :type show_full_error: bool
        :param return_json: Determines if the response should be returned in JSON format (defaults to ``True``)
        :returns: The API response in JSON format or as a ``requests`` object
        """
        return api.get(self, endpoint=endpoint, params=params, headers=headers, timeout=timeout,
                       show_full_error=show_full_error, return_json=return_json)

    def api_call_with_payload(self, method, endpoint, payload, params=None, headers=None, timeout=30,
                              show_full_error=True, return_json=True):
        """This method performs a POST call against the Salesforce instance.
        (`Reference <https://jereze.com/code/authentification-salesforce-rest-api-python/>`_)

        :param method: The API method (``post``, ``put``, or ``patch``)
        :type method: str
        :param endpoint: The API endpoint to query
        :type endpoint: str
        :param payload: The payload to leverage in the API call
        :type payload: dict
        :param params: The query parameters (where applicable)
        :type params: dict, None
        :param headers: Specific API headers to use when performing the API call
        :type headers: dict, None
        :param timeout: The timeout period in seconds (defaults to ``30``)
        :type timeout: int, str, None
        :param show_full_error: Determines if the full error message should be displayed (defaults to ``True``)
        :type show_full_error: bool
        :param return_json: Determines if the response should be returned in JSON format (defaults to ``True``)
        :returns: The API response in JSON format or as a ``requests`` object
        """
        return api.api_call_with_payload(self, method=method, endpoint=endpoint, payload=payload, params=params,
                                         headers=headers, timeout=timeout, show_full_error=show_full_error,
                                         return_json=return_json)

    def post(self, endpoint, payload, params=None, headers=None, timeout=30, show_full_error=True, return_json=True):
        """This method performs a POST call against the Salesforce instance.
        (`Reference <https://jereze.com/code/authentification-salesforce-rest-api-python/>`_)

        :param endpoint: The API endpoint to query
        :type endpoint: str
        :param payload: The payload to leverage in the API call
        :type payload: dict
        :param params: The query parameters (where applicable)
        :type params: dict, None
        :param headers: Specific API headers to use when performing the API call
        :type headers: dict, None
        :param timeout: The timeout period in seconds (defaults to ``30``)
        :type timeout: int, str, None
        :param show_full_error: Determines if the full error message should be displayed (defaults to ``True``)
        :type show_full_error: bool
        :param return_json: Determines if the response should be returned in JSON format (defaults to ``True``)
        :returns: The API response in JSON format or as a ``requests`` object
        """
        return api.api_call_with_payload(self, 'post', endpoint=endpoint, payload=payload, params=params,
                                         headers=headers, timeout=timeout, show_full_error=show_full_error,
                                         return_json=return_json)

    def patch(self, endpoint, payload, params=None, headers=None, timeout=30, show_full_error=True, return_json=False):
        """This method performs a PATCH call against the Salesforce instance.
        (`Reference <https://jereze.com/code/authentification-salesforce-rest-api-python/>`_)

        :param endpoint: The API endpoint to query
        :type endpoint: str
        :param payload: The payload to leverage in the API call
        :type payload: dict
        :param params: The query parameters (where applicable)
        :type params: dict, None
        :param headers: Specific API headers to use when performing the API call
        :type headers: dict, None
        :param timeout: The timeout period in seconds (defaults to ``30``)
        :type timeout: int, str, None
        :param show_full_error: Determines if the full error message should be displayed (defaults to ``True``)
        :type show_full_error: bool
        :param return_json: Determines if the response should be returned in JSON format (defaults to ``True``)
        :returns: The API response in JSON format or as a ``requests`` object
        """
        return api.api_call_with_payload(self, 'patch', endpoint=endpoint, payload=payload, params=params,
                                         headers=headers, timeout=timeout, show_full_error=show_full_error,
                                         return_json=return_json)

    def put(self, endpoint, payload, params=None, headers=None, timeout=30, show_full_error=True, return_json=True):
        """This method performs a PUT call against the Salesforce instance.
        (`Reference <https://jereze.com/code/authentification-salesforce-rest-api-python/>`_)

        :param endpoint: The API endpoint to query
        :type endpoint: str
        :param payload: The payload to leverage in the API call
        :type payload: dict
        :param params: The query parameters (where applicable)
        :type params: dict, None
        :param headers: Specific API headers to use when performing the API call
        :type headers: dict, None
        :param timeout: The timeout period in seconds (defaults to ``30``)
        :type timeout: int, str, None
        :param show_full_error: Determines if the full error message should be displayed (defaults to ``True``)
        :type show_full_error: bool
        :param return_json: Determines if the response should be returned in JSON format (defaults to ``True``)
        :returns: The API response in JSON format or as a ``requests`` object
        """
        return api.api_call_with_payload(self, 'put', endpoint=endpoint, payload=payload, params=params,
                                         headers=headers, timeout=timeout, show_full_error=show_full_error,
                                         return_json=return_json)

    def get_api_versions(self):
        """This method returns the API versions for the Salesforce releases.
        (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_versions.htm>`_)
        """
        return self.get('/services/data')

    def get_org_limits(self):
        """This method returns a list of all org limits.

        .. versionadded:: 1.1.0
        """
        return self.get(f'/services/data/{self.version}/limits')

    def get_all_sobjects(self):
        """This method returns a list of all Salesforce objects. (i.e. sObjects)
        (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_describeGlobal.htm>`_)
        """
        return self.get(f'/services/data/{self.version}/sobjects')

    def get_sobject(self, object_name, describe=False):
        """This method returns basic information or the full (describe) information for a specific sObject.
        (`Reference 1 <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_sobject_basic_info_get.htm>`_,
        `Reference 2 <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_sobject_describe.htm>`_)

        :param object_name: The name of the Salesforce object
        :type object_name: str
        :param describe: Determines if the full (i.e. ``describe``) data should be returned (defaults to ``False``)
        :type describe: bool
        :returns: The Salesforce object data
        """
        uri = f'/services/data/{self.version}/sobjects/{object_name}'
        uri = f'{uri}/describe' if describe else uri
        return self.get(uri)

    def describe_object(self, object_name):
        """This method returns the full (describe) information for a specific sObject.
        (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_sobject_describe.htm>`_)

        :param object_name: The name of the Salesforce object
        :type object_name: str
        :returns: The Salesforce object data
        """
        return self.get_sobject(object_name, describe=True)

    def get_rest_resources(self):
        """This method returns a list of all available REST resources.
        (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_discoveryresource.htm>`_)
        """
        return self.get(f'/services/data/{self.version}')

    def soql_query(self, query, replace_quotes=True, next_records_url=False):
        """This method performs a SOQL query and returns the results in JSON format.
        (`Reference 1 <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_query.htm>`_,
        `Reference 2 <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/knowledge_development_soql_sosl_intro.htm>`_)

        :param query: The SOQL query to perform
        :type query: str
        :param replace_quotes: Determines if double-quotes should be replaced with single-quotes (``True`` by default)
        :type replace_quotes: bool
        :param next_records_url: Indicates that the ``query`` parameter is a ``nextRecordsUrl`` value.
        :type next_records_url: bool
        :returns: The result of the SOQL query
        """
        if next_records_url:
            query = re.sub(r'^.*/', '', query) if '/' in query else query
        else:
            if replace_quotes:
                query = query.replace('"', "'")
            query = core_utils.url_encode(query)
            query = f'?q={query}'
        return self.get(f'/services/data/{self.version}/query/{query}')

    def search_string(self, string_to_search):
        """This method performs a SOSL query to search for a given string.
        (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_search.htm>`_)

        .. versionadded:: 1.1.0

        :param string_to_search: The string for which to search
        :type string_to_search: str
        :returns: The SOSL response data in JSON format
        """
        query = 'FIND {' + string_to_search + '}'
        query = core_utils.url_encode(query)
        return self.get(f'/services/data/{self.version}/search/?q={query}')


    def create_sobject_record(self, sobject, payload):
        """This method creates a new record for a specific sObject.
        (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_sobject_create.htm>`_)

        :param sobject: The sObject under which to create the new record
        :type sobject: str
        :param payload: The JSON payload with the record details
        :type payload: dict
        :returns: The API response from the POST request
        :raises: :py:exc:`RuntimeError`, :py:exc:`TypeError`
        """
        # Ensure the payload is in the appropriate format
        if not isinstance(payload, dict):
            raise TypeError('The sObject payload must be provided as a dictionary.')

        # Perform the API call and return the response
        response = self.post(f'/services/data/{self.version}/sobjects/{sobject}', payload=payload)
        return response

    def update_sobject_record(self, sobject, record_id, payload):
        """This method updates an existing sObject record.
        (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_update_fields.htm>`_)

        :param sobject: The sObject under which to update the record
        :type sobject: str
        :param record_id: The ID of the record to be updated
        :type record_id: str
        :param payload: The JSON payload with the record details to be updated
        :type payload: dict
        :returns: The API response from the PATCH request
        :raises: :py:exc:`RuntimeError`, :py:exc:`TypeError`
        """
        # Ensure the payload is in the appropriate format
        if not isinstance(payload, dict):
            raise TypeError('The sObject payload must be provided as a dictionary.')

        # Perform the API call and return the response
        response = self.patch(f'/services/data/{self.version}/sobjects/{sobject}/{record_id}', payload=payload)
        return response

    def download_image(self, image_url, record_id, field_name, file_path=None, sobject=None):
        """This method downloads an image using the sObject Rich Text Image Retrieve functionality.
        (`Reference 1 <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_sobject_rich_text_image_retrieve.htm>`_,
        `Reference 2 <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_sobject_rich_text_image_retrieve.htm>`_)

        :param image_url: The URL for the image to be downloaded
        :type image_url: str
        :param record_id: The Record ID where the image is found
        :type record_id: str
        :param field_name: The field name within the record where the image is found
        :type field_name: str
        :param file_path: The path to the directory where the image should be saved (current directory if not defined)
        :type file_path: str, None
        :param sobject: The sObject for the record where the image is found (``Knowledge__kav`` by default)
        :type sobject: str
        :returns: The full path to the downloaded image
        :raises: :py:exc:`RuntimeError`
        """
        # Ensure a valid sObject is defined (SFDC Knowledge unless otherwise specified)
        sobject = 'Knowledge__kav' if not sobject else sobject

        # Retrieve the reference ID for the image
        ref_id = core_utils.get_image_ref_id(image_url)

        # Define the URI and perform the API call
        image_path = None
        try:
            uri = f'/services/data/{self.version}/sobjects/{sobject}/{record_id}/richTextImageFields/{field_name}/{ref_id}'
            response = self.get(uri, return_json=False)

            # Save the image as an image file
            try:
                image_path = core_utils.download_image(file_name=f'{ref_id}.jpeg', file_path=file_path,
                                                       response=response)
            except RuntimeError:
                errors.handlers.eprint(f'Failed to download the image with refid {ref_id}.')
        except RuntimeError as exc:
            errors.handlers.eprint(exc)
        return image_path

    class Chatter(object):
        """This class includes methods associated with Salesforce Chatter."""
        def __init__(self, sfdc_object):
            """This method initializes the :py:class:`salespyforce.core.Salesforce.Chatter` inner class object.

            :param sfdc_object: The core :py:class:`salespyforce.Salesforce` object
            :type sfdc_object: class[salespyforce.Salesforce]
            """
            self.sfdc_object = sfdc_object

        def get_my_news_feed(self, site_id=None):
            """This method retrieves the news feed for the user calling the function.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.chatterapi.meta/chatterapi/quickreference_get_news_feed.htm>`_)

            :param site_id: The ID of an Experience Cloud site against which to query (optional)
            :type site_id: str, None
            :returns: The news feed data
            :raises: :py:exc:`RuntimeError`
            """
            return chatter_module.get_my_news_feed(self.sfdc_object, site_id=site_id)

        def get_user_news_feed(self, user_id, site_id=None):
            """This method retrieves another user's news feed.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.chatterapi.meta/chatterapi/quickreference_get_user_profile_feed.htm>`_)

            :param user_id: The ID of the user whose feed you wish to return
            :type user_id: str
            :param site_id: The ID of an Experience Cloud site against which to query (optional)
            :type site_id: str, None
            :returns: The news feed data
            :raises: :py:exc:`RuntimeError`
            """
            return chatter_module.get_user_news_feed(self.sfdc_object, user_id=user_id, site_id=site_id)

        def get_group_feed(self, group_id, site_id=None):
            """This method retrieves a group's news feed.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.chatterapi.meta/chatterapi/quickreference_get_group_feed.htm>`_)

            :param group_id: The ID of the group whose feed you wish to return
            :type group_id: str
            :param site_id: The ID of an Experience Cloud site against which to query (optional)
            :type site_id: str, None
            :returns: The news feed data
            :raises: :py:exc:`RuntimeError`
            """
            return chatter_module.get_group_feed(self.sfdc_object, group_id=group_id, site_id=site_id)

        def post_feed_item(self, subject_id, message_text=None, message_segments=None, site_id=None, created_by_id=None):
            """This method publishes a new Chatter feed item.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.chatterapi.meta/chatterapi/quickreference_post_feed_item.htm>`_)

            :param subject_id: The Subject ID against which to publish the feed item (e.g. ``0F9B000000000W2``)
            :type subject_id: str
            :param message_text: Plaintext to be used as the message body
            :type message_segments: str, None
            :param message_segments: Collection of message segments to use instead of a plaintext message
            :type message_segments: list, None
            :param site_id: The ID of an Experience Cloud site against which to query (optional)
            :type site_id: str, None
            :param created_by_id: The ID of the user to impersonate (**Experimental**)
            :type created_by_id: str, None
            :returns: The response of the POST request
            :raises: :py:exc:`RuntimeError`
            """
            return chatter_module.post_feed_item(self.sfdc_object, subject_id=subject_id, message_text=message_text,
                                                 message_segments=message_segments, site_id=site_id,
                                                 created_by_id=created_by_id)

        def post_comment(self, feed_element_id, message_text=None, message_segments=None, site_id=None, created_by_id=None):
            """This method publishes a comment on a Chatter feed item.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.chatterapi.meta/chatterapi/quickreference_post_comment_to_feed_element.htm>`_)

            :param feed_element_id: The ID of the feed element on which to post the comment
            :type feed_element_id: str
            :param message_text: Plaintext to be used as the message body
            :type message_segments: str, None
            :param message_segments: Collection of message segments to use instead of a plaintext message
            :type message_segments: list, None
            :param site_id: The ID of an Experience Cloud site against which to query (optional)
            :type site_id: str, None
            :param created_by_id: The ID of the user to impersonate (**Experimental**)
            :type created_by_id: str, None
            :returns: The response of the POST request
            :raises: :py:exc:`RuntimeError`
            """
            return chatter_module.post_comment(self.sfdc_object, feed_element_id=feed_element_id,
                                               message_text=message_text, message_segments=message_segments,
                                               site_id=site_id, created_by_id=created_by_id)

    class Knowledge(object):
        """This class includes methods associated with Salesforce Knowledge."""
        def __init__(self, sfdc_object):
            """This method initializes the :py:class:`salespyforce.core.Salesforce.Knowledge` inner class object.

            :param sfdc_object: The core :py:class:`salespyforce.Salesforce` object
            :type sfdc_object: class[salespyforce.Salesforce]
            """
            self.sfdc_object = sfdc_object

        def check_for_existing_article(self, title, sobject=None, return_id=False, return_id_and_number=False,
                                       include_archived=False):
            """This method checks to see if an article already exists with a given title and returns its article number.
            (`Reference 1 <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_query.htm>`_.
            `Reference 2 <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/knowledge_development_soql_sosl_intro.htm>`_)

            :param title: The title of the knowledge article for which to check
            :type title: str
            :param sobject: The Salesforce object to query (``Knowledge__kav`` by default)
            :type sobject: str, None
            :param return_id: Determines if the Article ID should be returned (``False`` by default)
            :type return_id: bool
            :param return_id_and_number: Determines if Article ID and Article Number should be returned (``False`` by default)
            :type return_id_and_number: bool
            :param include_archived: Determines if archived articles should be included (``False`` by default)
            :type include_archived: bool
            :returns: The Article Number, Article ID, or both, if found, or a blank string if not found
            """
            return knowledge_module.check_for_existing_article(self.sfdc_object, title=title,
                                                               sobject=sobject, return_id=return_id,
                                                               return_id_and_number=return_id_and_number,
                                                               include_archived=include_archived)

        def get_article_id_from_number(self, article_number, sobject=None, return_uri=False):
            """This method returns the Article ID when an article number is provided.
            (`Reference 1 <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_query.htm>`_,
            `Reference 2 <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/knowledge_development_soql_sosl_intro.htm>`_)

            :param article_number: The Article Number to query
            :type article_number: str, int
            :param sobject: The Salesforce object to query (``Knowledge__kav`` by default)
            :type sobject: str, None
            :param return_uri: Determines if the URI of the article should be returned rather than the ID (``False`` by default)
            :type return_uri: bool
            :returns: The Article ID or Article URI, or a blank string if no article is found
            :raises: :py:exc:`ValueError`
            """
            return knowledge_module.get_article_id_from_number(self.sfdc_object, article_number=article_number,
                                                               sobject=sobject, return_uri=return_uri)

        def get_articles_list(self, query=None, sort=None, order=None, page_size=20, page_num=1):
            """This method retrieves a list of knowledge articles.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_knowledge_support_artlist.htm>`_)

            :param query: A SOQL query with which to filter the results (optional)
            :type query: str, None
            :param sort: One of the following optional values: ``LastPublishedDate``, ``CreatedDate``, ``Title``, or ``ViewScore``
            :type sort: str, None
            :param order: Optionally define the ORDER BY as ``ASC`` or ``DESC``
            :type order: str, None
            :param page_size: The number of results per page (``20`` by default)
            :type page_size: int
            :param page_num: The starting page number (``1`` by default)
            :type page_num: int
            :returns: The list of retrieved knowledge articles
            """
            return knowledge_module.get_articles_list(self.sfdc_object, query=query, sort=sort, order=order,
                                                      page_size=page_size, page_num=page_num)

        def get_article_details(self, article_id, sobject=None):
            """This method retrieves details for a single knowledge article.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_knowledge_support_artdetails.htm>`_)

            :param article_id: The Article ID for which to retrieve details
            :type article_id: str
            :param sobject: The Salesforce object to query (``Knowledge__kav`` by default)
            :type sobject: str, None
            :returns: The details for the knowledge article
            """
            return knowledge_module.get_article_details(self.sfdc_object, article_id=article_id, sobject=sobject)

        def get_validation_status(self, article_id=None, article_details=None, sobject=None):
            """This method retrieves the Validation Status for a given Article ID.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_knowledge_support_artdetails.htm>`_)

            :param article_id: The Article ID for which to retrieve details
            :type article_id: str, None
            :param article_details: The dictionary of article details for the given article
            :type article_details: dict, None
            :param sobject: The Salesforce object to query (``Knowledge__kav`` by default)
            :type sobject: str, None
            :returns: The validation status as a text string
            :raises: :py:exc:`RuntimeError`
            """
            return knowledge_module.get_validation_status(self.sfdc_object, article_id=article_id,
                                                          article_details=article_details, sobject=sobject)

        def get_article_metadata(self, article_id):
            """This method retrieves metadata for a specific knowledge article.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/knowledge_REST_retrieve_article_metadata.htm>`_)

            :param article_id: The Article ID for which to retrieve details
            :type article_id: str
            :returns: The article metadata as a dictionary
            :raises: :py:exc:`RuntimeError`
            """
            return knowledge_module.get_article_metadata(self.sfdc_object, article_id=article_id)

        def get_article_version(self, article_id):
            """This method retrieves the version ID for a given master article ID.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/knowledge_REST_retrieve_article_version.htm>`_)

            :param article_id: The Article ID for which to retrieve details
            :type article_id: str
            :returns: The version ID for the given master article ID
            :raises: :py:exc:`RuntimeError`
            """
            return knowledge_module.get_article_version(self.sfdc_object, article_id=article_id)

        def get_article_url(self, article_id=None, article_number=None, sobject=None):
            """This function constructs the URL to view a knowledge article in Lightning or Classic.

            :param article_id: The Article ID for which to retrieve details
            :type article_id: str, None
            :param article_number: The article number for which to retrieve details
            :type article_number: str, int, None
            :param sobject: The Salesforce object to query (``Knowledge__kav`` by default)
            :type sobject: str, None
            :returns: The article URL as a string
            :raises: :py:exc:`ValueError`
            """
            return knowledge_module.get_article_url(self.sfdc_object, article_id=article_id,
                                                    article_number=article_number, sobject=sobject)

        def create_article(self, article_data, sobject=None, full_response=False):
            """This method creates a new knowledge article draft.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_sobject_create.htm>`_)

            :param article_data: The article data used to populate the article
            :type article_data: dict
            :param sobject: The Salesforce object to query (``Knowledge__kav`` by default)
            :type sobject: str, None
            :param full_response: Determines if the full API response should be returned instead of the article ID (``False`` by default)
            :type full_response: bool
            :returns: The API response or the ID of the article draft
            :raises: :py:exc:`ValueError`, :py:exc:`TypeError`
            """
            return knowledge_module.create_article(self.sfdc_object, article_data=article_data, sobject=sobject,
                                                   full_response=full_response)

        def update_article(self, record_id, article_data, sobject=None, include_status_code=False):
            """This method updates an existing knowledge article draft.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_update_fields.htm>`_)

            :param record_id: The ID of the article draft record to be updated
            :type record_id: str
            :param article_data: The article data used to update the article
            :type article_data: dict
            :param sobject: The Salesforce object to query (``Knowledge__kav`` by default)
            :type sobject: str, None
            :param include_status_code: Determines if the API response status code should be returned (``False`` by default)
            :type include_status_code: bool
            :returns: A Boolean indicating if the update operation was successful, and optionally the API response status code
            :raises: :py:exc:`ValueError`, :py:exc:`TypeError`, :py:exc:`RuntimeError`
            """
            return knowledge_module.update_article(self.sfdc_object, record_id=record_id, article_data=article_data,
                                                   sobject=sobject, include_status_code=include_status_code)

        def create_draft_from_online_article(self, article_id, unpublish=False):
            """This method creates a draft knowledge article from an online article.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/actions_obj_knowledge.htm#createDraftFromOnlineKnowledgeArticle>`_)

            :param article_id: The ID of the online article from which to create the draft
            :type article_id: str
            :param unpublish: Determines if the online article should be unpublished when the draft is created (``False`` by default)
            :type unpublish: bool
            :returns: The API response from the POST request
            :raises: :py:exc:`RuntimeError`
            """
            return knowledge_module.create_draft_from_online_article(self.sfdc_object, article_id=article_id,
                                                                     unpublish=unpublish)

        def create_draft_from_master_version(self, article_id=None, knowledge_article_id=None, article_data=None,
                                             sobject=None, full_response=False):
            """This method creates an online version of a master article.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.198.0.knowledge_dev.meta/knowledge_dev/knowledge_REST_edit_online_master.htm>`_)

            :param article_id: The Article ID from which to create the draft
            :type article_id: str, None
            :param knowledge_article_id: The Knowledge Article ID (``KnowledgeArticleId``) from which to create the draft
            :type knowledge_article_id: str, None
            :param article_data: The article data associated with the article from which to create the draft
            :type article_data: dict, None
            :param sobject: The Salesforce object to query (``Knowledge__kav`` by default)
            :type sobject: str, None
            :param full_response: Determines if the full API response should be returned instead of the article ID (``False`` by default)
            :type full_response: bool
            :returns: The API response or the ID of the article draft
            :raises: :py:exc:`RuntimeError`
            """
            return knowledge_module.create_draft_from_master_version(self.sfdc_object, article_id=article_id,
                                                                     knowledge_article_id=knowledge_article_id,
                                                                     article_data=article_data, sobject=sobject,
                                                                     full_response=full_response)

        def publish_article(self, article_id, major_version=True, full_response=False):
            """This method publishes a draft knowledge article as a major or minor version.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/knowledge_REST_publish_master_version.htm>`_)

            :param article_id: The Article ID to publish
            :type article_id: str
            :param major_version: Determines if the published article should be a major version (``True`` by default)
            :type major_version: bool
            :param full_response: Determines if the full API response should be returned (``False`` by default)
            :type full_response: bool
            :returns: A Boolean value indicating the success of the action or the API response from the PATCH request
            :raises: :py:exc:`RuntimeError`
            """
            return knowledge_module.publish_article(self.sfdc_object, article_id=article_id,
                                                    major_version=major_version, full_response=full_response)

        def publish_multiple_articles(self, article_id_list, major_version=True):
            """This method publishes multiple knowledge article drafts at one time.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/actions_obj_knowledge.htm#publishKnowledgeArticles>`_)

            :param article_id_list: A list of Article IDs to be published
            :type article_id_list: list
            :param major_version: Determines if the published article should be a major version (``True`` by default)
            :type major_version: bool
            :returns: The API response from the POST request
            :raises: :py:exc:`RuntimeError`, :py:exc:`TypeError`, :py:exc:`ValueError`
            """
            return knowledge_module.publish_multiple_articles(self.sfdc_object, article_id_list=article_id_list,
                                                              major_version=major_version)

        def assign_data_category(self, article_id, category_group_name, category_name):
            """This method assigns a single data category for a knowledge article.
            (`Reference <https://itsmemohit.medium.com/quick-win-15-salesforce-knowledge-rest-apis-bb0725b2040e>`_)

            .. versionadded:: 1.2.0

            :param article_id: The ID of the article to update
            :type article_id: str
            :param category_group_name: The unique Data Category Group Name
            :type category_group_name: str
            :param category_name: The unique Data Category Name
            :type category_name: str
            :returns: The API response from the POST request
            :raises: :py:exc:`RuntimeError`
            """
            return knowledge_module.assign_data_category(self.sfdc_object, article_id=article_id,
                                                         category_group_name=category_group_name,
                                                         category_name=category_name)

        def archive_article(self, article_id):
            """This function archives a published knowledge article.
            (`Reference <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/knowledge_REST_archive_master_version.htm>`_)

            .. versionadded:: 1.3.0

            :param article_id: The ID of the article to archive
            :type article_id: str
            :returns: The API response from the POST request
            :raises: :py:exc:`RuntimeError`
            """
            return knowledge_module.archive_article(self.sfdc_object, article_id=article_id)


def define_connection_info():
    """This function prompts the user for the connection information.

    :returns: The connection info in a dictionary
    """
    base_url = input('Enter your instance URL: [] ')
    org_id = input('Enter the Org ID for your instance: [] ')
    username = input('Enter the username of your API user: [] ')
    password = input('Enter the password of your API user: [] ')
    endpoint_url = input('Enter the endpoint URL: [] ')
    client_id = input('Enter the Client ID: [] ')
    client_secret = input('Enter the Client Secret: [] ')
    security_token = input('Enter the Security Token: [] ')
    connection_info = compile_connection_info(base_url, org_id, username, password, endpoint_url,
                                              client_id, client_secret, security_token)
    return connection_info


def compile_connection_info(base_url, org_id, username, password, endpoint_url,
                            client_id, client_secret, security_token):
    """This function compiles the connection info into a dictionary that can be consumed by the core object.

    :param base_url: The base URL of the Salesforce instance
    :type base_url: str
    :param org_id: The Org ID of the Salesforce instance
    :type org_id: str
    :param username: The username of the API user
    :type username: str
    :param password: The password of the API user
    :type password: str
    :param endpoint_url: The endpoint URL for the Salesforce instance
    :type endpoint_url: str
    :param client_id: The Client ID for the Salesforce instance
    :type client_id: str
    :param client_secret: The Client Secret for the Salesforce instance
    :type client_secret: str
    :param security_token: The Security Token for the Salesforce instance
    :type security_token: str
    :returns: The connection info in a dictionary
    """
    connection_info = {
        'base_url': base_url,
        'org_id': org_id,
        'username': username,
        'password': password,
        'endpoint_url': endpoint_url,
        'client_id': client_id,
        'client_secret': client_secret,
        'security_token': security_token,
    }
    return connection_info



# -*- coding: utf-8 -*-
"""
:Package:        salespyforce.utils
:Synopsis:       This is the ``__init__`` module for the salespyforce.utils modules
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  13 Mar 2023
"""

__all__ = ['core_utils', 'helper', 'version']



# -*- coding: utf-8 -*-
"""
:Module:            salespyforce.knowledge
:Synopsis:          Defines the Knowledge-related functions associated with the Salesforce API
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     14 Nov 2023
"""

from . import errors
from .utils import log_utils

# Initialize logging
logger = log_utils.initialize_logging(__name__)


def check_for_existing_article(sfdc_object, title, sobject=None, return_id=False, return_id_and_number=False,
                               include_archived=False):
    """This method checks to see if an article already exists with a given title and returns its article number.
    (`Reference 1 <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_query.htm>`_,
    `Reference 2 <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/knowledge_development_soql_sosl_intro.htm>`_)

    .. versionchanged:: 1.2.2
       You can now specify whether archived articles are included in the query results.

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param title: The title of the knowledge article for which to check
    :type title: str
    :param sobject: The Salesforce object to query (``Knowledge__kav`` by default)
    :type sobject: str, None
    :param return_id: Determines if the Article ID should be returned (``False`` by default)
    :type return_id: bool
    :param return_id_and_number: Determines if Article ID and Article Number should be returned (``False`` by default)
    :type return_id_and_number: bool
    :param include_archived: Determines if archived articles should be included (``False`` by default)
    :type include_archived: bool
    :returns: The Article Number, Article ID, or both, if found, or a blank string if not found
    """
    sobject = 'Knowledge__kav' if sobject is None else sobject
    query = f"SELECT Id,ArticleNumber FROM {sobject} WHERE Title = '{title}'"
    query += " AND PublishStatus != 'Archived'" if not include_archived else query
    response = sfdc_object.soql_query(query, replace_quotes=False)
    return_value = ''
    if response.get('totalSize') > 0:
        if return_id:
            return_value = response['records'][0]['Id']
        elif return_id_and_number:
            return_value = (response['records'][0]['Id'], response['records'][0]['ArticleNumber'])
        else:
            return_value = response['records'][0]['ArticleNumber']
    elif return_id_and_number:
        return_value = ('', '')
    return return_value


def get_article_id_from_number(sfdc_object, article_number, sobject=None, return_uri=False):
    """This method returns the Article ID when an article number is provided.
    (`Reference 1 <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_query.htm>`_,
    `Reference 2 <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/knowledge_development_soql_sosl_intro.htm>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param article_number: The Article Number to query
    :type article_number: str, int
    :param sobject: The Salesforce object to query (``Knowledge__kav`` by default)
    :type sobject: str, None
    :param return_uri: Determines if the URI of the article should be returned rather than the ID (``False`` by default)
    :type return_uri: bool
    :returns: The Article ID or Article URI, or a blank string if no article is found
    :raises: :py:exc:`ValueError`
    """
    sobject = 'Knowledge__kav' if sobject is None else sobject
    if sobject is None:
        raise ValueError('The sObject must be defined for the Article Type in order to query for the ID.')
    if len(str(article_number)) < 9:
        query = f"SELECT Id FROM {sobject} WHERE ArticleNumber LIKE '%0{article_number}'"
    else:
        query = f"SELECT Id FROM {sobject} WHERE ArticleNumber = '{article_number}'"
    response = sfdc_object.soql_query(query)
    if response.get('totalSize') > 0:
        if return_uri:
            return_value = response['records'][0]['attributes']['url']
        else:
            return_value = response['records'][0]['Id']
    else:
        return_value = ''
        print(f'No results were returned when querying for the article number {article_number}')
    return return_value


def get_articles_list(sfdc_object, query=None, sort=None, order=None, page_size=20, page_num=1):
    """This function retrieves a list of knowledge articles.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_knowledge_support_artlist.htm>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param query: A SOQL query with which to filter the results (optional)
    :type query: str, None
    :param sort: One of the following optional values: ``LastPublishedDate``, ``CreatedDate``, ``Title``, or ``ViewScore``
    :type sort: str, None
    :param order: Optionally define the ORDER BY as ``ASC`` or ``DESC``
    :type order: str, None
    :param page_size: The number of results per page (``20`` by default)
    :type page_size: int
    :param page_num: The starting page number (``1`` by default)
    :type page_num: int
    :returns: The list of retrieved knowledge articles
    """
    # Define the headers
    headers = sfdc_object._get_headers('articles')

    # Validate the sort field
    valid_sort_options = ['LastPublishedDate', 'CreatedDate', 'Title', 'ViewScore']
    if sort and sort not in valid_sort_options:
        errors.handlers.eprint(f'The sort value {sort} is not valid and will be ignored.')
        sort = None

    # Validate the order field
    if order and order.upper() not in ['ASC', 'DESC']:
        errors.handlers.eprint(f'The order value {order} is not valid and will be ignored.')
        order = None

    # Validate the page size field
    if page_size > 100:
        errors.handlers.eprint(f'The pageSize value exceeds the maximum and will default to 100.')
        page_size = 100

    # Validate the pageNumber field
    if page_num < 1:
        errors.handlers.eprint(f'The pageNumber value is not valid and will default to 1.')
        page_num = 1

    # Add values to the parameters dictionary if they have been defined
    params = {}
    if query:
        params['q'] = query
    if sort:
        params['sort'] = sort
    if order:
        params['order'] = order
    params['pageSize'] = page_size
    params['pageNumber'] = page_num

    # Perform the query
    return sfdc_object.get(f'/services/data/{sfdc_object.version}/support/knowledgeArticles',
                           params=params, headers=headers)


def get_article_details(sfdc_object, article_id, sobject=None):
    """This function retrieves details for a single knowledge article.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_knowledge_support_artdetails.htm>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param article_id: The Article ID for which to retrieve details
    :type article_id: str
    :param sobject: The Salesforce object to query (``Knowledge__kav`` by default)
    :type sobject: str, None
    :returns: The details for the knowledge article
    """
    # Define the headers
    headers = sfdc_object._get_headers('articles')

    # Perform the query and return the data
    sobject = 'Knowledge__kav' if sobject is None else sobject
    if sobject is not None:
        data = sfdc_object.get(f'/services/data/{sfdc_object.version}/sobjects/{sobject}/{article_id}')
    else:
        data = sfdc_object.get(f'/services/data/{sfdc_object.version}/support/knowledgeArticles/{article_id}',
                               headers=headers)
    return data


def get_validation_status(sfdc_object, article_id=None, article_details=None, sobject=None):
    """This function retrieves the Validation Status for a given Article ID.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_knowledge_support_artdetails.htm>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param article_id: The Article ID for which to retrieve details
    :type article_id: str, None
    :param article_details: The dictionary of article details for the given article
    :type article_details: dict, None
    :param sobject: The Salesforce object to query (``Knowledge__kav`` by default)
    :type sobject: str, None
    :returns: The validation status as a text string
    :raises: :py:exc:`RuntimeError`
    """
    if not any((article_id, article_details)):
        raise RuntimeError('The article ID or article details must be provided.')

    # Retrieve the article details if not already supplied
    if not article_details:
        article_details = get_article_details(sfdc_object, article_id, sobject)

    # Identify the validation status
    return article_details.get('ValidationStatus')


def get_article_metadata(sfdc_object, article_id):
    """This function retrieves metadata for a specific knowledge article.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/knowledge_REST_retrieve_article_metadata.htm>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param article_id: The Article ID for which to retrieve details
    :type article_id: str
    :returns: The article metadata as a dictionary
    :raises: :py:exc:`RuntimeError`
    """
    return sfdc_object.get(f'/services/data/{sfdc_object.version}/knowledgeManagement/articles/{article_id}')


def get_article_version(sfdc_object, article_id):
    """This function retrieves the version ID for a given master article ID.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/knowledge_REST_retrieve_article_version.htm>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param article_id: The Article ID for which to retrieve details
    :type article_id: str
    :returns: The version ID for the given master article ID
    :raises: :py:exc:`RuntimeError`
    """
    endpoint = f'/services/data/{sfdc_object.version}/knowledgeManagement/articleversions/masterVersions/{article_id}'
    return sfdc_object.get(endpoint)


def get_article_url(sfdc_object, article_id=None, article_number=None, sobject=None):
    """This function constructs the URL to view a knowledge article in Lightning or Classic.

    .. versionchanged:: 1.2.0
       Changed when lightning URLs are defined and fixed an issue with extraneous slashes.

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param article_id: The Article ID for which to retrieve details
    :type article_id: str, None
    :param article_number: The article number for which to retrieve details
    :type article_number: str, int, None
    :param sobject: The Salesforce object to query (``Knowledge__kav`` by default)
    :type sobject: str, None
    :returns: The article URL as a string
    :raises: :py:exc:`ValueError`
    """
    sobject = 'Knowledge__kav' if sobject is None else sobject
    if not any((article_id, article_number)):
        raise ValueError('An article ID or an article number must be provided to retrieve the article URL.')
    if article_number and not article_id:
        article_id = get_article_id_from_number(sfdc_object, article_number, sobject)
    segment = '' if sfdc_object.base_url.endswith('/') else '/'
    if 'lightning' in sfdc_object.base_url or sobject == 'Knowledge__kav':
        article_url = f'{sfdc_object.base_url}{segment}lightning/r/Knowledge__kav/{article_id}/view'
    else:
        article_url = f'{sfdc_object.base_url}{segment}knowledge/publishing/articleDraftDetail.apexp?id={article_id}'
    return article_url


def create_article(sfdc_object, article_data, sobject=None, full_response=False):
    """This function creates a new knowledge article draft.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_sobject_create.htm>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param article_data: The article data used to populate the article
    :type article_data: dict
    :param sobject: The Salesforce object to query (``Knowledge__kav`` by default)
    :type sobject: str, None
    :param full_response: Determines if the full API response should be returned instead of the article ID (``False`` by default)
    :type full_response: bool
    :returns: The API response or the ID of the article draft
    :raises: :py:exc:`ValueError`, :py:exc:`TypeError`, :py:exc:`RuntimeError`
    """
    # Get the appropriate sObject to call
    sobject = 'Knowledge__kav' if sobject is None else sobject

    # Ensure the payload is in the appropriate format
    if not isinstance(article_data, dict):
        raise TypeError('The article data must be provided as a dictionary.')

    # Ensure that the required fields have been provided
    required_fields = ['Title', 'UrlName']
    for field in required_fields:
        if field not in article_data:
            raise ValueError(f'The following required field is missing from the article data: {field}')

    # Perform the API call
    response = sfdc_object.post(f'/services/data/{sfdc_object.version}/sobjects/{sobject}', payload=article_data)

    # Return the full response or just the article ID
    if not full_response:
        response = response.get('id')
    return response


def update_article(sfdc_object, record_id, article_data, sobject=None, include_status_code=False):
    """This function updates an existing knowledge article draft.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_update_fields.htm>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param record_id: The ID of the article draft record to be updated
    :type record_id: str
    :param article_data: The article data used to update the article
    :type article_data: dict
    :param sobject: The Salesforce object to query (``Knowledge__kav`` by default)
    :type sobject: str, None
    :param include_status_code: Determines if the API response status code should be returned (``False`` by default)
    :type include_status_code: bool
    :returns: A Boolean indicating if the update operation was successful, and optionally the API response status code
    :raises: :py:exc:`ValueError`, :py:exc:`TypeError`, :py:exc:`RuntimeError`
    """
    # Get the appropriate sObject to call
    sobject = 'Knowledge__kav' if sobject is None else sobject

    # Ensure the payload is in the appropriate format
    if not isinstance(article_data, dict):
        raise TypeError('The article data must be provided as a dictionary.')

    # Ensure that the required fields have been provided
    required_fields = ['Title', 'UrlName']
    for field in required_fields:
        if field not in article_data:
            raise ValueError(f'The following required field is missing from the article data: {field}')

    # Perform the API call
    response = sfdc_object.patch(f'/services/data/{sfdc_object.version}/sobjects/{sobject}/{record_id}',
                                 payload=article_data)

    # Determine whether the call was successful
    successful = True if response.status_code == 204 else False

    # Return the success determination and optionally the status code
    if include_status_code:
        return successful, response.status_code
    return successful


def create_draft_from_online_article(sfdc_object, article_id, unpublish=False):
    """This function creates a draft knowledge article from an online article.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/actions_obj_knowledge.htm#createDraftFromOnlineKnowledgeArticle>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param article_id: The ID of the online article from which to create the draft
    :type article_id: str
    :param unpublish: Determines if the online article should be unpublished when the draft is created (``False`` by default)
    :type unpublish: bool
    :returns: The API response from the POST request
    :raises: :py:exc:`RuntimeError`
    """
    # Define the payload for the API call
    payload = {
        "inputs": [
            {
                "action": "EDIT_AS_DRAFT_ARTICLE",
                "unpublish": unpublish,
                "articleId": f"{article_id}"
            }
        ]
    }

    # Perform the API call
    endpoint = f'/services/data/{sfdc_object.version}/actions/standard/createDraftFromOnlineKnowledgeArticle'
    return sfdc_object.post(endpoint, payload)


def create_draft_from_master_version(sfdc_object, article_id=None, knowledge_article_id=None, article_data=None,
                                     sobject=None, full_response=False):
    """This function creates an online version of a master article.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.198.0.knowledge_dev.meta/knowledge_dev/knowledge_REST_edit_online_master.htm>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param article_id: The Article ID from which to create the draft
    :type article_id: str, None
    :param knowledge_article_id: The Knowledge Article ID (``KnowledgeArticleId``) from which to create the draft
    :type knowledge_article_id: str, None
    :param article_data: The article data associated with the article from which to create the draft
    :type article_data: dict, None
    :param sobject: The Salesforce object to query (``Knowledge__kav`` by default)
    :type sobject: str, None
    :param full_response: Determines if the full API response should be returned instead of the article ID (``False`` by default)
    :type full_response: bool
    :returns: The API response or the ID of the article draft
    :raises: :py:exc:`RuntimeError`
    """
    if not any((article_id, knowledge_article_id, article_data)):
        raise RuntimeError('Need to provide article ID, knowledge article ID, or article data.')

    # Get the appropriate sObject to call
    sobject = 'Knowledge__kav' if sobject is None else sobject

    # Get the knowledge article ID as needed
    if not knowledge_article_id:
        if not article_data:
            article_data = sfdc_object.get_article_details(article_id, sobject=sobject)
        knowledge_article_id = article_data.get('KnowledgeArticleId')

    # Perform the API call to retrieve the new draft ID
    endpoint = f'/services/data/{sfdc_object.version}/knowledgeManagement/articleVersions/masterVersions'
    response = sfdc_object.post(endpoint, {'articleId': knowledge_article_id})

    # Return the full response or the draft ID
    if not full_response:
        response = response.get('id')
    return response


def publish_article(sfdc_object, article_id, major_version=True, full_response=False):
    """This function publishes a draft knowledge article as a major or minor version.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/knowledge_REST_publish_master_version.htm>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param article_id: The Article ID to publish
    :type article_id: str
    :param major_version: Determines if the published article should be a major version (``True`` by default)
    :type major_version: bool
    :param full_response: Determines if the full API response should be returned (``False`` by default)
    :type full_response: bool
    :returns: A Boolean value indicating the success of the action or the API response from the PATCH request
    :raises: :py:exc:`RuntimeError`
    """
    # Define the payload for the API call
    payload = {
        "publishStatus": "Online"
    }
    if major_version:
        payload['versionNumber'] = 'NextVersion'

    # Perform the API call
    endpoint = f'/services/data/{sfdc_object.version}/knowledgeManagement/articleVersions/masterVersions/{article_id}'
    result = sfdc_object.patch(endpoint, payload)

    # Return the appropriate value depending on if a full response was requested
    if not full_response:
        result = True if result.status_code == 204 else False
    return result


def publish_multiple_articles(sfdc_object, article_id_list, major_version=True):
    """This function publishes multiple knowledge article drafts at one time.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/actions_obj_knowledge.htm#publishKnowledgeArticles>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param article_id_list: A list of Article IDs to be published
    :type article_id_list: list
    :param major_version: Determines if the published article should be a major version (``True`` by default)
    :type major_version: bool
    :returns: The API response from the POST request
    :raises: :py:exc:`RuntimeError`, :py:exc:`TypeError`, :py:exc:`ValueError`
    """
    # Define the endpoint URI
    endpoint = f'/services/data/{sfdc_object.version}/actions/standard/publishKnowledgeArticles'

    # Ensure there is at least one article ID to publish
    if not isinstance(article_id_list, list) or not isinstance(article_id_list[0], str):
        raise TypeError('A list of Article ID strings must be provided in order to publish multiple articles.')
    elif len(article_id_list) == 0:
        raise ValueError('No article ID strings were found in the article ID list variable.')

    # Define the action to perform
    action = 'PUBLISH_ARTICLE_NEW_VERSION' if major_version else 'PUBLISH_ARTICLE'

    # Construct the payload
    payload = {
        "inputs": [
            {
                "articleVersionIdList": article_id_list,
                "pubAction": action
            }
        ]
    }

    # Perform the API call
    return sfdc_object.post(endpoint, payload)


def assign_data_category(sfdc_object, article_id, category_group_name, category_name):
    """This function assigns a single data category for a knowledge article.
    (`Reference <https://itsmemohit.medium.com/quick-win-15-salesforce-knowledge-rest-apis-bb0725b2040e>`_)

    .. versionadded:: 1.2.0

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param article_id: The ID of the article to update
    :type article_id: str
    :param category_group_name: The unique Data Category Group Name
    :type category_group_name: str
    :param category_name: The unique Data Category Name
    :type category_name: str
    :returns: The API response from the POST request
    :raises: :py:exc:`RuntimeError`
    """
    # Define the payload for the API call
    payload = {
        "ParentId": article_id,
        "DataCategoryGroupName": category_group_name,
        "DataCategoryName": category_name
    }

    # Perform the API call
    endpoint = f'/services/data/{sfdc_object.version}/sobjects/Knowledge__DataCategorySelection'
    return sfdc_object.post(endpoint, payload)


def archive_article(sfdc_object, article_id):
    """This function archives a published knowledge article.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.knowledge_dev.meta/knowledge_dev/knowledge_REST_archive_master_version.htm>`_)

    .. versionadded:: 1.3.0

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param article_id: The ID of the article to archive
    :type article_id: str
    :returns: The API response from the POST request
    :raises: :py:exc:`RuntimeError`
    """
    # Define the payload for the API call
    payload = {
        "publishStatus": "Archived"
    }

    # Perform the API call
    endpoint = f'/services/data/{sfdc_object.version}/knowledgeManagement/articleVersions/masterVersions/{article_id}'
    return sfdc_object.patch(endpoint, payload)



# -*- coding: utf-8 -*-
"""
:Module:         salespyforce.utils.tests.test_sobjects
:Synopsis:       This module is used by pytest to test performing SOQL queries
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  29 May 2023
"""

from . import resources


def test_soql_query():
    """This function tests the ability to perform a SOQL query.

    .. versionadded:: 1.1.0
    """
    sfdc_object = resources.get_core_object()
    soql_statement = 'SELECT Id FROM Account LIMIT 1'
    soql_response = sfdc_object.soql_query(soql_statement)
    assert 'done' in soql_response and soql_response.get('done') is True
    assert 'totalSize' in soql_response and 'records' in soql_response




# -*- coding: utf-8 -*-
"""
:Module:            salespyforce.utils.tests.resources
:Synopsis:          Frequently used resources for performing unit testing
:Usage:             ``from salespyforce.utils.tests import resources``
:Example:           ``exceptions = resources.import_exceptions_module()``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     27 May 2023
"""

import os
import sys
import importlib

import pytest

# Define constants
SKIP_LOCAL_TEST_MSG = 'skipping local-only tests'
HELPER_FILE_NAME = 'helper_shurt.yml'


class MockResponse:
    """This class simulates an API response for testing purposes.

    .. versionadded:: 1.1.0
    """
    def __init__(self, json_body, status_code=200):
        self.json_body = json_body
        self.status_code = status_code

    def json(self):
        return self.json_body


def mock_success_post(*args, **kwargs):
    """This function works with the `MockedResponse` class to simulate a successful API response.

    .. versionadded:: 1.1.0
    """
    return MockResponse({
        "id": "001D000000IqhSLIAZ",
        "errors": [],
        "success": True
    })


def mock_error_post(*args, **kwargs):
    """This function works with the `MockedResponse` class to simulate a failed API response.

    .. versionadded:: 1.1.0
    """
    return MockResponse({
        "errors": [],
        "success": False
    })


def mock_sosl_get(*args, **kwargs):
    """This function works with the `MockedResponse` class to simulate an SOSL response."""
    return MockResponse({
        "searchRecords": [
            {
                'attributes': {
                    'type': 'Account',
                    'url': '/services/data/v57.0/sobjects/Account/0018V00002NeqAxQAJ'
                },
                'Id': '0018V00002NeqAxQAJ'
            }
        ]
    })


def set_package_path():
    """This function adds the high-level salespyforce directory to the sys.path list.

    .. versionadded:: 1.1.0
    """
    sys.path.insert(0, os.path.abspath('../..'))


def import_modules(*modules):
    """This function imports and returns one or more modules to utilize in a unit test.

    .. versionadded:: 1.1.0

    :param modules: One or more module paths (absolute) in string format
    :returns: The imported module(s) as an individual object or a tuple of objects
    """
    imported_modules = []
    for module in modules:
        imported_modules.append(importlib.import_module(module))
    tuple(imported_modules)
    return imported_modules if len(imported_modules) > 1 else imported_modules[0]


def secrets_helper_exists():
    """This function checks to see if the unencrypted helper file exists for GitHub Actions.

    .. versionadded:: 1.1.0
    """
    helper_path = f'{os.environ.get("HOME")}/secrets/{HELPER_FILE_NAME}'
    return os.path.isfile(helper_path)


def local_helper_exists():
    """This function checks to see if a helper file is present in the ``local/`` directory.

    .. versionadded:: 1.1.0
    """
    return os.path.exists(f'local/{HELPER_FILE_NAME}')


def get_core_object():
    """This function instantiates and returns the core object using a local helper file.

    .. versionadded:: 1.1.0
    """
    set_package_path()
    if secrets_helper_exists():
        sfdc_object = instantiate_with_secrets_helper()
    else:
        if not local_helper_exists():
            pytest.skip('skipping tests where a valid helper file is needed')
        sfdc_object = instantiate_with_local_helper()
    return sfdc_object


def instantiate_with_secrets_helper():
    """This function instantiates the Salesforce object using the unencrypted helper file intended for GitHub Actions.

    .. versionadded:: 1.1.0

    :returns: The instantiated :py:class:`salespyforce.core.Salesforce` object
    :raises: :py:exc:`FileNotFoundError`
    """
    if not secrets_helper_exists():
        raise FileNotFoundError('The unencrypted GitHub Actions helper file cannot be found.')
    file_name = f'{os.environ.get("HOME")}/secrets/{HELPER_FILE_NAME}'
    set_package_path()
    core_module = importlib.import_module('salespyforce.core')
    return core_module.Salesforce(helper=file_name)


def instantiate_with_local_helper():
    """This function instantiates the Salesforce object using a local helper file for unit testing.

    .. versionadded:: 1.1.0

    :returns: The instantiated :py:class:`salespyforce.core.Salesforce` object
    :raises: :py:exc:`FileNotFoundError`
    """
    if not local_helper_exists():
        raise FileNotFoundError('The local helper file cannot be found.')
    set_package_path()
    core_module = importlib.import_module('salespyforce.core')
    return core_module.Salesforce(helper=f"local/{HELPER_FILE_NAME}")



# -*- coding: utf-8 -*-
"""
:Module:            salespyforce.utils.core_utils
:Synopsis:          Collection of supporting utilities and functions to complement the primary modules
:Usage:             ``from salespyforce.utils import core_utils``
:Example:           ``encoded_string = core_utils.encode_url(decoded_string)``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     29 May 2023
"""

import random
import string
import os.path
import warnings
import urllib.parse

import requests

from . import log_utils
from .. import errors

# Initialize the logger for this module
logger = log_utils.initialize_logging(__name__)


def url_encode(raw_string):
    """This function encodes a string for use in URLs.

    :param raw_string: The raw string to be encoded
    :type raw_string: str
    :returns: The encoded string
    """
    return urllib.parse.quote_plus(raw_string)


def url_decode(encoded_string):
    """This function decodes a url-encoded string.

    :param encoded_string: The url-encoded string
    :type encoded_string: str
    :returns: The unencoded string
    """
    return urllib.parse.unquote_plus(encoded_string)


def display_warning(warn_msg):
    """This function displays a :py:exc:`UserWarning` message via the :py:mod:`warnings` module.

    :param warn_msg: The message to be displayed
    :type warn_msg: str
    :returns: None
    """
    warnings.warn(warn_msg, UserWarning)


def get_file_type(file_path):
    """This function attempts to identify if a given file path is for a YAML or JSON file.

    :param file_path: The full path to the file
    :type file_path: str
    :returns: The file type in string format (e.g. ``yaml`` or ``json``)
    :raises: :py:exc:`FileNotFoundError`, :py:exc:`khoros.errors.exceptions.UnknownFileTypeError`
    """
    file_type = 'unknown'
    if os.path.isfile(file_path):
        if file_path.endswith('.json'):
            file_type = 'json'
        elif file_path.endswith('.yml') or file_path.endswith('.yaml'):
            file_type = 'yaml'
        else:
            display_warning(f"Unable to recognize the file type of '{file_path}' by its extension.")
            with open(file_path) as cfg_file:
                for line in cfg_file:
                    if line.startswith('#'):
                        continue
                    else:
                        if '{' in line:
                            file_type = 'json'
                            break
        if file_type == 'unknown':
            raise errors.exceptions.UnknownFileTypeError(file=file_path)
    else:
        raise FileNotFoundError(f"Unable to locate the following file: {file_path}")
    return file_type


def get_random_string(length=32, prefix_string=""):
    """This function returns a random alphanumeric string.

    :param length: The length of the string (``32`` by default)
    :type length: int
    :param prefix_string: A string to which the random string should be appended (optional)
    :type prefix_string: str
    :returns: The alphanumeric string
    """
    return f"{prefix_string}{''.join([random.choice(string.ascii_letters + string.digits) for _ in range(length)])}"


def get_image_ref_id(image_url):
    """This function parses an image URL to identify the reference ID (refid) value.
    (`Reference 1 <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_sobject_rich_text_image_retrieve.htm>`_,
    `Reference 2 <https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_sobject_rich_text_image_retrieve.htm>`_)

    :param image_url: The URL of an image from within Salesforce
    :type image_url: str
    :returns: The reference ID (``refid``) value
    """
    query_params = urllib.parse.parse_qs(urllib.parse.urlparse(image_url).query)
    # noinspection PyTypeChecker
    ref_id = query_params.get('refid')
    ref_id = ref_id[0] if not isinstance(ref_id, str) else ref_id
    return ref_id


def download_image(image_url=None, file_name=None, file_path=None, response=None, extension='jpeg'):
    """This function downloads an image and saves it to a specified directory.

    :param image_url: The absolute URL to the image
    :type image_url: str
    :param file_name: The file name including extension as which to save the file
    :type file_name: str
    :param file_path: File path where the image file should be saved (default: ``var/images/``)
    :type file_path: str, None
    :param response: The response of the previously performed API call
    :param extension: The file extension to use if a file name with extension is not provided
    :type extension: str
    :returns: The full path to the downloaded image
    :raises: :py:exc:`RuntimeError`
    """
    if not image_url and not response:
        raise RuntimeError('An image URL or an API response must be provided to download an image.')

    # Define an appropriate file path
    file_path = './' if not file_path else file_path
    file_path = f'{file_path}/' if not any((file_path.endswith('/'), file_path.endswith('\\'))) else file_path

    # Define a file name if not provided
    if not file_name:
        file_name = get_random_string(10, 'image_')
        file_name += extension

    # Perform the API call if not supplied
    if not response:
        response = requests.get(image_url)
    if response.status_code != 200:
        raise RuntimeError(f'The image failed to download with a {response.status_code} status code.')

    # Export the response data as an image file
    with open(f'{file_path}{file_name}', 'wb') as file:
        file.write(response.content)
    return f'{file_path}{file_name}'



# -*- coding: utf-8 -*-
"""
:Module:         salespyforce.utils.tests
:Synopsis:       This package includes tests for the salespyforce library.
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  27 May 2023
"""



# -*- coding: utf-8 -*-
"""
:Module:            salespyforce.chatter
:Synopsis:          Defines the Chatter-related functions associated with the Salesforce Connect API
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     13 Mar 2023
"""

from .utils import log_utils

# Initialize logging
logger = log_utils.initialize_logging(__name__)


def _get_site_endpoint_segment(_site_id=None):
    """This function constructs the endpoint segment when querying a specific Experience Cloud site.

    :param _site_id: The Site ID of the Experience Cloud site
    :type _site_id: str
    :returns: The API endpoint segment (or a blank string if no Site ID was provided)
    """
    _endpoint_segment = f'/connect/communities/{_site_id}' if _site_id else ''
    return _endpoint_segment


def get_my_news_feed(sfdc_object, site_id=None):
    """This function retrieves the news feed for the user calling the function.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.chatterapi.meta/chatterapi/quickreference_get_news_feed.htm>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param site_id: The ID of an Experience Cloud site against which to query (optional)
    :type site_id: str, None
    :returns: The news feed data
    :raises: :py:exc:`RuntimeError`
    """
    site_segment = _get_site_endpoint_segment(site_id)
    endpoint = f'/services/data/{sfdc_object.version}{site_segment}/chatter/feeds/news/me/feed-elements'
    return sfdc_object.get(endpoint)


def get_user_news_feed(sfdc_object, user_id, site_id=None):
    """This function retrieves another user's news feed.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.chatterapi.meta/chatterapi/quickreference_get_user_profile_feed.htm>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param user_id: The ID of the user whose feed you wish to return
    :type user_id: str
    :param site_id: The ID of an Experience Cloud site against which to query (optional)
    :type site_id: str, None
    :returns: The news feed data
    :raises: :py:exc:`RuntimeError`
    """
    site_segment = _get_site_endpoint_segment(site_id)
    endpoint = f'/services/data/{sfdc_object.version}{site_segment}/chatter/feeds/user-profile/{user_id}/feed-elements'
    return sfdc_object.get(endpoint)


def get_group_feed(sfdc_object, group_id, site_id=None):
    """This function retrieves a group's news feed.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.chatterapi.meta/chatterapi/quickreference_get_group_feed.htm>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param group_id: The ID of the group whose feed you wish to return
    :type group_id: str
    :param site_id: The ID of an Experience Cloud site against which to query (optional)
    :type site_id: str, None
    :returns: The news feed data
    :raises: :py:exc:`RuntimeError`
    """
    site_segment = _get_site_endpoint_segment(site_id)
    endpoint = f'/services/data/{sfdc_object.version}{site_segment}/chatter/feeds/record/{group_id}/feed-elements'
    return sfdc_object.get(endpoint)


def post_feed_item(sfdc_object, subject_id, message_text=None, message_segments=None, site_id=None, created_by_id=None):
    """This function publishes a new Chatter feed item.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.chatterapi.meta/chatterapi/quickreference_post_feed_item.htm>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param subject_id: The Subject ID against which to publish the feed item (e.g. ``0F9B000000000W2``)
    :type subject_id: str
    :param message_text: Plaintext to be used as the message body
    :type message_segments: str, None
    :param message_segments: Collection of message segments to use instead of a plaintext message
    :type message_segments: list, None
    :param site_id: The ID of an Experience Cloud site against which to query (optional)
    :type site_id: str, None
    :param created_by_id: The ID of the user to impersonate (**Experimental**)
    :type created_by_id: str, None
    :returns: The response of the POST request
    :raises: :py:exc:`RuntimeError`
    """
    site_segment = _get_site_endpoint_segment(site_id)
    if not any((message_text, message_segments)):
        raise RuntimeError('Message text or message segments are required to post a feed item.')
    if not message_segments:
        message_segments = _construct_simple_message_segment(message_text)
    body = {
        'body': {
            'messageSegments': message_segments
        },
        'feedElementType': 'FeedItem',
        'subjectId': subject_id,
    }
    if created_by_id:
        body['createdById'] = created_by_id
    endpoint = f'/services/data/{sfdc_object.version}{site_segment}/chatter/feed-elements?' \
               f'feedElementType=FeedItem&subjectId={subject_id}'
    return sfdc_object.post(endpoint=endpoint, payload=body)


def post_comment(sfdc_object, feed_element_id, message_text=None, message_segments=None, site_id=None, created_by_id=None):
    """This function publishes a comment on a Chatter feed item.
    (`Reference <https://developer.salesforce.com/docs/atlas.en-us.chatterapi.meta/chatterapi/quickreference_post_comment_to_feed_element.htm>`_)

    :param sfdc_object: The instantiated SalesPyForce object
    :type sfdc_object: class[salespyforce.Salesforce]
    :param feed_element_id: The ID of the feed element on which to post the comment
    :type feed_element_id: str
    :param message_text: Plaintext to be used as the message body
    :type message_segments: str, None
    :param message_segments: Collection of message segments to use instead of a plaintext message
    :type message_segments: list, None
    :param site_id: The ID of an Experience Cloud site against which to query (optional)
    :type site_id: str, None
    :param created_by_id: The ID of the user to impersonate (**Experimental**)
    :type created_by_id: str, None
    :returns: The response of the POST request
    :raises: :py:exc:`RuntimeError`
    """
    site_segment = _get_site_endpoint_segment(site_id)
    if not any((message_text, message_segments)):
        raise RuntimeError('Message text or message segments are required to post a feed comment.')
    if not message_segments:
        message_segments = _construct_simple_message_segment(message_text)
    body = {
        'body': {
            'messageSegments': message_segments
        }
    }
    if created_by_id:
        body['createdById'] = created_by_id
    endpoint = f'/services/data/{sfdc_object.version}{site_segment}/chatter/feed-elements/' \
               f'{feed_element_id}/capabilities/comments/items'
    return sfdc_object.post(endpoint=endpoint, payload=body)


def _construct_simple_message_segment(_message_text):
    """This function constructs a simple message segments collection to be used in an API payload.

    :param _message_text: The plaintext message to be embedded in a message segment.
    :type _message_text: str
    :returns: The constructed message segments payload
    """
    _message_segments = [
        {
            'type': 'text',
            'text': _message_text
        }
    ]
    return _message_segments

